@@ -307,7 +307,7 @@ class WizataDSAPIClient(ApiInterface):
             method="POST",
             route="execute/data",
             result_type="pickle",
-            data=json.dumps(request.prepare(), cls=DSAPIEncoder)
+            data=json.dumps(request.to_json(), cls=DSAPIEncoder)
         )
 
     def info(self):
@@ -645,7 +645,7 @@ class WizataDSAPIClient(ApiInterface):
         request = Request()
 
         if datapoints is not None:
-            request.add_datapoints(datapoints)
+            request.datapoints = datapoints
 
         if start is not None:
             request.start = start
@@ -806,7 +806,7 @@ class WizataDSAPIClient(ApiInterface):
 
         response = requests.request("POST", self.__url() + "data/string",
                                     headers=self.__header(),
-                                    data=json.dumps(request.prepare(), cls=DSAPIEncoder))
+                                    data=json.dumps(request.to_json(), cls=DSAPIEncoder))
         if response.status_code == 200:
             return response.json()
         else:
